
# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: src/ai/backend/plugin:dist

from setuptools import setup

setup(**{
    'author': 'Lablup Inc. and contributors',
    'classifiers': [
        'Intended Audience :: Developers',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Environment :: No Input/Output (Daemon)',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development',
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3.13',
        'License :: OSI Approved :: MIT License',
    ],
    'description': 'Backend.AI Plugin Subsystem',
    'entry_points': {
        'backendai_cli_v10': [
            'plugin = ai.backend.plugin.cli:main',
        ],
    },
    'install_requires': (
        'backend.ai-logging==25.6.12rc1',
        'click~=8.1.7',
        'colorama>=0.4.6',
        'tabulate~=0.8.9',
        'types-tabulate',
    ),
    'license': 'MIT',
    'long_description': """Backend.AI Plugin Subsystem
===========================

Package Structure
-----------------

* `ai.backend.plugin`: Abstract types for plugins and a common base plugin set
""",
    'long_description_content_type': 'text/markdown',
    'name': 'backend.ai-plugin',
    'namespace_packages': (
    ),
    'package_data': {
        'ai.backend.plugin': (
            'VERSION',
            'py.typed',
        ),
    },
    'packages': (
        'ai.backend.plugin',
    ),
    'project_urls': {
        'Documentation': 'https://docs.backend.ai/',
        'Source': 'https://github.com/lablup/backend.ai',
    },
    'python_requires': '>=3.13,<3.14',
    'url': 'https://github.com/lablup/backend.ai',
    'version': '25.6.12rc1',
    'zip_safe': False,
})
